



# 内核编译

## 环境安装
选题方向：模块实验设计

以 Ubuntu20.04 为例

```shell
# 编译语言：
sudo apt-get install gcc g++ make cmake libelf-dev libc6-dev libglib2.0-dev libidn11 libidn11-dev libncurses-dev libncurses5-dev libncursesw5-dev libpixman-1-dev libssl-dev libtool

# 内核挂载（模块化所需）：
sudo apt-get install zstd dwarves autoconf build-essential bison flex software-properties-common xz-utils kernel-package openssl zlibc minizip ncurses-dev

# 原生F2FS格式化工具：
sudo apt-get install uuid-dev pkg-config autoconf automake libtool f2fs-tools
# 或者下载源码编译 https://git.kernel.org/pub/scm/linux/kernel/git/jaegeuk/f2fs-tools.git 这里默认是1.14.0

# 其他（包括duperemove需要）：
sudo apt-get install snappy*  bzip2* lzo* zlib*  lz4* gzip* openssl* ncurses* liblz4-tool sqlite3 libdb-dev fio
```

## 编译

```shell
wget https://mirror.bjtu.edu.cn/kernel/linux/kernel/v5.x/linux-5.10.50.tar.xz
# 解压进入源码根目录
```

```shell
sudo make menuconfig
# File systems		--->
# 	<M>	F2FS filesystem support
# 	[ * ]		F2FS Status Information
# 	[ * ]		F2FS extended attributes
# 	[ * ]			F2FS Access Control Lists
# 	[ * ]			F2FS Security Labels
# 	[   ]		F2FS consistency checking feature
# 	[   ]		F2FS Encryption
# 	[   ]		F2FS IO tracer
# 	[   ]		F2FS fault injection facility
# XFS 和 BTRFS 默认勾选，最好选上二者的 debug 选项
```

```shell
scripts/config --disable SYSTEM_TRUSTED_KEYS
scripts/config --disable SYSTEM_REVOCATION_KEYS

sudo make -j$(nproc --all)
make INSTALL_MOD_STRIP=1 modules_install
sudo make install
```

## 挂载

编译后修改开机启动项选择进入5.10.50

F2DFS/F2FS 源码包下Makefile修改，指向对应源码目录

```
KERNEL_DIR:=.../KERNEL/fs/f2fs
PWD:=.../KERNEL/fs/f2fs

default:
	make -C $(KERNEL_DIR) M=$(PWD) modules

clean:
	rm -rf *.o *.mod.c *.ko *.symvers modules.order
```

然后 `make clean & make` 即可（这里F2DFS模块名依旧是f2fs.ko）

对于F2FS:

```shell
insmod f2fs.ko
mkdir -p /mnt/f2fs
mkfs.f2fs /dev/sda
mount -t f2fs /dev/sda /mnt/f2fs
# ...
umount /mnt/f2fs
rmmod f2fs.ko
```


对于F2DFS:

```shell
insmod f2fs.ko
mkdir -p /mnt/f2dfs
# 编译f2dfs-tools目录下的格式化工具
.../f2dfs-tools/mkfs/mkfs.f2fs /dev/sdb
mount -t f2fs /dev/sdb /mnt/f2dfs
# ...
umount /mnt/f2dfs
rmmod f2fs.ko
```

## 测试样例

```shell
fio -filename=/mnt/.../test_fio -buffered=1 -dedupe_percentage=50 -iodepth 1 -rw=write -ioengine=sync -bs=1M -size=1G -numjobs=4 -group_reporting -name=XXXX
# buffered=1 buffered写会在线去重，direct=1 direct写，直达磁盘
# dedupe_percentage：重复率设置
# rw：write/read/randwrite/randread
# ioengine：选用引擎，direct/buffered同步引擎均sync，direct异步引擎libaio，buffered异步引擎io_uring
# bs：块大小
# size：文件总大小，最终落盘数据量
# numjobs：测试线程数，总共 size * numjobs 大小的写入
```


# 其他


## 其他 F2DFS 参数设置

```shell
# 查看当前 F2DFS 状态信息
car /sys/kernel/debug/f2fs/status
```

```c
// f2dfs_fs.h
// 哈希指纹长度，例如 SHA-1 20字节，SHA-256 32字节
#define FP_LEN_MAX						(20)
// 指纹前缀长度，默认2^14个并行桶
#define FP_SUFFIX_LENGTH_BIT			(4)
```

```c
// f2dfs_fp.h
// 最大并行哈希数，这个数字与CPU消耗无关，只是用于防止锁争用
#define F2DFS_HASH_JOBS_NUMBER			(16384)
// 桶内可以存放内存键值对数量
#define FP_ENTRIES_PER_BUFFER			(6000)
```

```c
// f2dfs_fingerprint.c

// 选择使用的哈希算法，默认是SHA-1
sbi->s_fingerprint_hash->basic_hash_info = &sha1_info;

// 如果要替换算法（例如 MD5，SHA256，trg192，mumur3 等），需要自行修改下面信息：
struct f2dfs_basic_hash_info sha1_info = {
  /* 必须为内核已经开启支持的哈希算法，比如 MD5 需要编译内核时打开 */
  .name =			"sha1",
  /* 这两个参数无用，只要不超过 struct f2dfs_basic_hash_info 限制字符串长度即可 */
	.simple_name =	"SHA-1",
	.full_name =	"Secure Hash Algorithm",
  /* 替换的哈希算法长度，来自上面的f2dfs_fs.h */
	.len =			FP_LEN_MAX,		/* 20 B */
  /*
   * 4KB大小全零块哈希结果
   * dd if=/dev/zero of=./zero_block bs=4k count=1
   * https://www.strerr.com/cn/sha1_file.html
   * 将生成的零块文件拖入网页进行哈希计算，相似的有 SHA256、MD5 等
   */
	.zero_result =	"\x1c\xea\xf7\x3d"		/* "1CEA":  7402  "F73D": 63293 */
					"\xf4\x0e\x53\x1d"		/* "F40E": 62478  "531D": 21277 */
					"\xf3\xbf\xb2\x6b"		/* "F3BF": 62399  "B26B": 45675 */
					"\x4f\xb7\xcd\x95"		/* "4FB7": 20407  "CD95": 52629 */
					"\xfb\x7b\xff\x1d",		/* "FB7B": 64379  "FF1D": 65309 */
  /* 全零块哈希结果的前两字节（unsigned int类型） */
	.zero_prefix =	0x1cea,
  /* 空结果，长度必须为哈希算法长度，即20个\x00 */
	.null_result =	"\x00\x00\x00\x00"		/* "0000":     0  "0000":     0 */
					"\x00\x00\x00\x00"		/* "0000":     0  "0000":     0 */
					"\x00\x00\x00\x00"		/* "0000":     0  "0000":     0 */
					"\x00\x00\x00\x00"		/* "0000":     0  "0000":     0 */
					"\x00\x00\x00\x00",		/* "0000":     0  "0000":     0 */
  /* 调用的哈希函数 */
  .fp_update =	crypto_shash_update,
};
```


