﻿#ifndef _F2DFS_FP_H_
#define _F2DFS_FP_H_

#include <linux/fs.h>
#include <crypto/hash.h>
#include <linux/buffer_head.h>
#include <linux/mpage.h>
#include <linux/writeback.h>
#include <linux/backing-dev.h>
#include <linux/pagevec.h>
#include <linux/blkdev.h>
#include <linux/bio.h>
#include <linux/blk-crypto.h>
#include <linux/swap.h>
#include <linux/prefetch.h>
#include <linux/uio.h>
#include <linux/cleancache.h>
#include <linux/sched/signal.h>
#include <linux/fiemap.h>
#include <linux/posix_acl_xattr.h>
#include <linux/lzo.h>
#include <linux/lz4.h>
#include <linux/zstd.h>
#include <linux/types.h>
#include <linux/page-flags.h>
#include <linux/slab.h>
#include <linux/crc32.h>
#include <linux/magic.h>
#include <linux/kobject.h>
#include <linux/sched.h>
#include <linux/cred.h>
#include <linux/vmalloc.h>
#include <linux/quotaops.h>
#include <linux/part_stat.h>
#include <linux/fscrypt.h>
#include <linux/fsverity.h>
#include <linux/stat.h>
#include <linux/falloc.h>
#include <linux/compat.h>
#include <linux/uaccess.h>
#include <linux/mount.h>
#include <linux/pagevec.h>
#include <linux/uuid.h>
#include <linux/file.h>
#include <linux/nls.h>
#include <linux/module.h>
#include <linux/backing-dev.h>
#include <linux/init.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/freezer.h>
#include <linux/pagemap.h>
#include <linux/unicode.h>
#include <linux/ctype.h>
#include <linux/random.h>
#include <linux/dcache.h>
#include <linux/namei.h>
#include <linux/timer.h>
#include <linux/freezer.h>
#include <linux/statfs.h>
#include <linux/parser.h>
#include <linux/mount.h>
#include <linux/seq_file.h>
#include <linux/proc_fs.h>
#include <linux/exportfs.h>
#include <linux/sysfs.h>
#include <linux/quota.h>
#include <linux/compiler.h>
#include <linux/radix-tree.h>
#include <linux/rwsem.h>
#include <linux/security.h>
#include <linux/xattr.h>
#include <trace/events/f2fs.h>
#include <uapi/linux/f2fs.h>
#include "f2fs.h"
#include "f2dfs_fs.h"

// 1. basic data structure

#define F2DFS_IOC_OFFLINE_DEDUPLICATION										\
				_IOW((F2FS_IOCTL_MAGIC), (20), __u32)
#define F2DFS_IOC_RELEASE_CACHE												\
				_IOW((F2FS_IOCTL_MAGIC), (21), __u32)

#define F2DFS_NOT_PRINT_DEDUPE_INFO		(0)
#define F2DFS_NOT_PRINT_DEDUPE_TRACE	(1)
#define F2DFS_NOT_PRINT_IO_TRACE		(1)

void f2dfs_printk(struct f2fs_sb_info *sbi, const char *fmt, ...);

#define F2DFS_HASH_JOBS_NUMBER			(16384)

#define FP_COUNTER_ZERO					(5)
#define FP_COUNTER_UNIQUE				(FP_COUNTER_ZERO + (1))
#define FP_COUNTER_DUPL					(FP_COUNTER_ZERO + (2))
#define PERSIST_NIDS_NUM				(FP_PREFIX_NUMBER * (2))
#define F2DFS_OFFLINE_RA_PAGES_NR		(BIO_MAX_PAGES)
#define SSA_VIRTUAL_ADDR				(0x8000)		// ofs in SSA of virtual

#define FP_COUNTER_MAX					(0x00ffffff)
#define VAT_FP_MASK_1					(0x0000ffff)	// FP prefix and stable
#define VAT_FP_MASK_2					(0xffff0000)	// FP prefix and stable
#define VAT_FP_MASK_3					(0x00007fff)	// FP prefix only
#define VAT_STABLE_MASK_1				(0x80000000)	// stable tag only
#define VAT_STABLE_MASK_2				(0x00008000)	// stable tag only
#define VAT_C_MASK_H_1					(0x0000ffff)
#define VAT_C_MASK_H_2					(0x00ffff00)
#define VAT_C_MASK_L					(0x000000ff)
#define DIRECT_FP_PREFIX				(FP_PREFIX_NUMBER)
#define INVALID_FP_PREFIX				(FP_PREFIX_NUMBER + 1)
#define ZERO_FP_PREFIX					(((FP_PREFIX_MASK_BIT) &			\
				(sbi->s_fingerprint_hash->basic_hash_info->zero_prefix))	\
				>> (((FP_PREFIX_LENGTH_BYTE) << (3)) - FP_PREFIX_LENGTH_BIT))
#define F2DFS_STABLE_BITMAP_SIZE		(DIV_ROUND_UP(FP_PREFIX_NUMBER, (32)))

struct slist_head {
	struct slist_head * next;
};

struct sdesc {
	struct shash_desc shash;
	char ctx[];
};

struct f2dfs_basic_hash_info {
	char * name;
	char * simple_name;
	char * full_name;
	unsigned long len;
	unsigned int zero_prefix;
	char * zero_result;
	char * null_result;
	int (*fp_update)(struct shash_desc *shash,
					const u8 *data, unsigned int len);


	struct crypto_shash *alg;
};

// Constants related to bucket and buffer configurations
#define FP_ENTRIES_PER_BUCKET            (160)
#define FP_ENTRIES_PER_BUFFER            (6000)
#define FP_ENTRIES_PER_SKIP              (32)
#define FP_REMOVE_PER_ENTRY              (FP_LENGTH_ADDR + (1)) // Remove address count
#define FP_BUCKET_EVICT_START_1          ((FP_ENTRIES_PER_BUCKET * (3)) >> (2)) // Eviction start threshold 1
#define FP_BUCKET_EVICT_START_2          ((FP_ENTRIES_PER_BUCKET) >> (1)) // Eviction start threshold 2
#define FP_BUCKET_EVICT_END              ((FP_ENTRIES_PER_BUCKET) >> (2)) // Eviction end threshold

// Structure representing changes in a bucket
struct f2dfs_bucket_change {
    __u8 fingerprint[FP_LEN_MAX];   // Fingerprint for deduplication
    virtual_t virtual_addr;         // Virtual address of the entry
    unsigned int fp_prefix;         // Prefix of fingerprint
    bool is_dupl;                   // Whether entry is a duplicate
    bool is_stable;
    bool is_persist;
    bool is_remove;
};

// Fingerprint list with associated pages and allocations
struct f2dfs_fp_list {
    struct page **pages;
    __u8 nr_pages; 
    __le16 index; 
    __u8 alloc[FP_LIST_MAX_LEN]; 
};

// Journal for managing fingerprint changes
struct f2dfs_fp_journal {
    struct page **pages; 
    int nr_pages; 
    nid_t next_nid;    
    __u8 alloc[FP_JOURNAL_MAX_LEN]; 
};

// Represents a normal fingerprint entry
struct fp_normal_entry {
    __u8 fingerprint[FP_LEN_MAX]; 
    virtual_t virtual_addr;  
};

// Represents an entry marked for removal
struct fp_remove_entry {
    virtual_t remove_addr[FP_REMOVE_PER_ENTRY];
};

// Represents a bucket entry with tags
struct fp_bucket_entry {
    union {
        struct fp_normal_entry ne; 
        struct fp_remove_entry re;
    };
    unsigned char tags;   
};

// Represents a fingerprint bucket with metadata
struct f2dfs_fp_bucket {
    struct slist_head b_head;
    struct fp_bucket_entry entries[FP_ENTRIES_PER_BUCKET];
    __u8 remove_count;             
    __u8 valid_count;                
    unsigned long atime;             
};

// Tags used for marking bucket entry states
#define FP_PERSIST_TAG          (0x01)
#define FP_REMOVE_TAG           (0x02)
#define FP_STABLE_TAG           (0x04)  
#define FP_DUPL_TAG             (0x08) 

// Inline functions for tag manipulation
static inline int is_fp_tag(struct fp_bucket_entry *entry, int type) {
    return ((entry->tags) & (type));    // Check if a tag is set
}

static inline void set_fp_tag(struct fp_bucket_entry *entry, int type) {
    ((entry->tags) |= (type));          // Set a specific tag
}

static inline void clear_fp_tag(struct fp_bucket_entry *entry, int type) {
    ((entry->tags) &= (~(type)));       // Clear a specific tag
}

static inline void init_fp_tag(struct fp_bucket_entry *entry) {
    ((entry->tags) &= (0xf0));          // Initialize tags to zero
}

// Tag checks and operations
#define fp_is_persist_tag(entry)        is_fp_tag((entry), FP_PERSIST_TAG)
#define fp_set_persist_tag(entry)       set_fp_tag((entry), FP_PERSIST_TAG)
#define fp_clear_persisit_tag(entry)    clear_fp_tag((entry), FP_PERSIST_TAG)

// Structures and methods for fingerprint hash jobs

// Represents a single hash job
struct f2dfs_single_hash_job {
    struct f2fs_sb_info *sbi; 
    struct work_struct page_work; 
    struct mutex work_mutex;    
    struct page *page;             
    nid_t ino;                   
    pgoff_t index;                 
    struct f2dfs_bucket_change search_result; 
    struct sdesc *sdesc;            
};

// Represents the fingerprint hash state
struct f2dfs_fingerprint_hash {
    struct f2dfs_basic_hash_info *basic_hash_info;
    struct workqueue_struct *parallel_hash_workqueue;
    struct f2dfs_single_hash_job *hash_jobs;        
};

// Represents fingerprint buffer metadata
struct f2dfs_fingerprint_buffer {
    unsigned int current_number;    
    struct kmem_cache *buffer_cache;
    struct slist_head *pointers;     
};

// 3. return result

enum {		/* cache replacement algorithm */
	FP_BUCKET_EVICTION_LRU = 0,			/* non scan-resistant */
	FP_BUCKET_EVICTION_FIFO,			/* Belay */
	FP_BUCKET_EVICTION_LFU,
	FP_BUCKET_EVICTION_ARC,				/* LRU + LFU */
	FP_BUCKET_EVICTION_CLOCK,
	FP_BUCKET_EVICTION_LRU_K,			/* scan-resistant */
	FP_BUCKET_EVICTION_TWO_QUEUES,		/* scan-resistant, LRU-2 */
	FP_BUCKET_EVICTION_LIRS,
	FP_BUCKET_EVICTION_DLRU,
	FP_BUCKET_EVICTION_DARC,
};

#define FP_BUCKET_EVICTION_CRA					(FP_BUCKET_EVICTION_LRU)

enum {		/* in-line deduplication */
	FP_HASH_SUCCESS = 0,
	FP_HASH_FAILURE,

	FP_BUCKET_CACHE_FOR_NOTHING,
	FP_BUCKET_CACHE_FOR_SEARCH,
	FP_BUCKET_CACHE_FOR_INSERT,
	FP_BUCKET_CACHE_FOR_REMOVE,
	FP_BUCKET_CACHE_FOR_GET_STABLE,
	FP_BUCKET_CACHE_FOR_SET_STABLE,

	FP_BUCKET_SEARCH_DUPL,
	FP_BUCKET_SEARCH_ZERO,
	FP_BUCKET_SEARCH_UNIQUE,
	FP_BUCKET_INSERT_SUCCESS,
	FP_BUCKET_INSERT_FAILURE,
	FP_BUCKET_REMOVE_SUCCESS,
	FP_BUCKET_REMOVE_FAILURE,

	FP_BUFFER_APPEND_NID_SUCCESS,
	FP_BUFFER_APPEND_NID_FAILURE,
	FP_BUFFER_SUBTRACT_NID_SUCCESS,
	FP_BUFFER_SUBTRACT_NID_FAILURE,
	FP_BUFFER_LOAD_NID_SUCCESS,
	FP_BUFFER_LOAD_NID_FAILURE,

	FP_BUCKET_SEARCH_STABLE,
	FP_BUCKET_SEARCH_UNSTABLE,
	FP_BUCKET_STABLE_SUCCESS,
	FP_BUCKET_STABLE_FAILURE,
};

enum {		/* persistence */
	FP_LIST_RETURN_NOTHING = 0,

	FP_LIST_SEARCH_SUCCESS,
	FP_LIST_SEARCH_FAILURE,
	FP_LIST_INSERT_SUCCESS,
	FP_LIST_INSERT_FAILURE,
	FP_LIST_REMOVE_SUCCESS,
	FP_LIST_REMOVE_FAILURE,
	FP_LIST_NODE_SEARCH_SUCCESS,
	FP_LIST_NODE_SEARCH_FAILURE,
	FP_LIST_NODE_INSERT_SUCCESS,
	FP_LIST_NODE_INSERT_FAILURE,
	FP_LIST_NODE_INSERT_IS_EXIST,
	FP_LIST_NODE_REMOVE_SUCCESS,
	FP_LIST_NODE_REMOVE_FAILURE,
	FP_LIST_NODE_ALLOC_SUCCESS,
	FP_LIST_NODE_ALLOC_FAILURE,
	FP_LIST_NODE_FREE_SUCCESS,
	FP_LIST_NODE_FREE_FAILURE,
	FP_LIST_NODE_IS_STABLE,
	FP_LIST_NODE_IS_UNSTABLE,
	FP_LIST_CONVERT_STABLE_SUCCESS,
	FP_LIST_CONVERT_STABLE_FAILURE,
	FP_LIST_IS_STABLE,
	FP_LIST_IS_UNSTABLE,
};

enum {		/* out-of-line deduplication */
	FP_JOURNAL_RETURN_NOTHING = 0,
	FP_JOURNAL_INSERT_SUCCESS,
	FP_JOURNAL_INSERT_FAILURE,
	FP_JOURNAL_REMOVE_SUCCESS,
	FP_JOURNAL_REMOVE_FAILURE,
	FP_JOURNAL_NODE_INSERT_SUCCESS,
	FP_JOURNAL_NODE_INSERT_FAILURE,
	FP_JOURNAL_NODE_REMOVE_SUCCESS,
	FP_JOURNAL_NODE_REMOVE_FAILURE,
	FP_JOURNAL_NODE_ALLOC_SUCCESS,
	FP_JOURNAL_NODE_ALLOC_FAILURE,
	FP_JOURNAL_NODE_FREE_SUCCESS,
	FP_JOURNAL_NODE_FREE_FAILURE,
};


// 4. VAT

#define vat_get_vaddr(nat)				((nat)->ni.nid)
#define vat_set_vaddr(nat, n)			((nat)->ni.nid = (n))
#define vat_get_blkaddr(nat)			((nat)->ni.blk_addr)
#define vat_set_blkaddr(nat, b)			((nat)->ni.blk_addr = (b))
#define vat_get_ino(nat)				((nat)->ni.ino)
#define vat_set_ino(nat, i)				((nat)->ni.ino = (i))
#define vat_get_version(nat)			((nat)->ni.version)
#define vat_set_version(nat, v)			((nat)->ni.version = (v))
#define vat_s_counter_low(c)			((c) & (VAT_C_MASK_L))
#define vat_s_counter_high(c)												\
				((((c) & (VAT_C_MASK_H_2)) >> (8)) & (VAT_C_MASK_H_1))
#define vat_s_fp(fp, tag)				(((((fp) & (VAT_FP_MASK_3)) +		\
				(((tag) == true) ? (VAT_STABLE_MASK_2) : (0))) &			\
				(VAT_FP_MASK_1)) << (16))
#define vat_g_counter_low(v)			((v) & (VAT_C_MASK_L))
#define vat_g_counter_high(i)												\
				((((i) & (VAT_C_MASK_H_1)) << (8)) & (VAT_C_MASK_H_2))
#define vat_g_fp(i)						((i) & (VAT_FP_MASK_2))
#define get_prefix_direct(i)			((((vat_g_fp(i)) >> (16)) &			\
				(VAT_FP_MASK_1)) & (VAT_FP_MASK_3))
#define get_prefix_stable(i)			(((vat_g_fp(i))) & (VAT_STABLE_MASK_1))
#define vat_get_prefix(nat)				(get_prefix_direct(vat_get_ino((nat))))
#define vat_set_prefix(nat, fp, tag)										\
				vat_set_ino((nat), (vat_g_counter_high(						\
				vat_get_ino((nat))) + vat_s_fp((fp), (tag))))
#define get_counter_direct(i, v)											\
				(vat_g_counter_high((i)) + vat_g_counter_low((v)))
#define vat_get_counter(nat)			vat_g_counter_high(					\
				vat_get_ino((nat))) + vat_g_counter_low(vat_get_version((nat)))
#define vat_set_counter(nat, c)												\
				vat_set_version((nat), vat_s_counter_low((c)));				\
				vat_set_ino((nat),											\
				(vat_g_fp(vat_get_ino((nat))) + vat_s_counter_high((c))));
#define vat_init_counter(nat, fp)											\
				vat_set_counter((nat), (FP_COUNTER_UNIQUE));				\
				vat_set_prefix((nat), (fp), (false));
#define vat_release_counter(nat)		vat_set_ino((nat), (0));			\
				vat_set_version((nat), (0));
#define is_vat_stable(i)													\
				((get_prefix_stable((i)) == (0)) ? (false) : (true))


// 5. offline and physical bitmap

#define get_dnode_nooffline(addr)											\
				((__is_valid_data_blkaddr(addr) && ((addr) != ZERO_ADDR)) ?	\
				(le32_to_cpu(addr) & (~(ADDRS_OFFLINE_TAG))) :				\
				le32_to_cpu(addr))

/* treat ZERO/NULL/NEW/COMPRESS/FSYNC/DIRECT ADDRS as fake physical, return 0 */
#define is_real_dnode_physical(addr)										\
				((!__is_valid_data_blkaddr(addr) || ((addr) != ZERO_ADDR)) ?\
				((ADDRS_PHYSICAL_TAG) & (addr)) : (0))
/* treat ZERO/NULL/NEW/COMPRESS/FSYNC/DIRECT ADDRS as physical, return 1 */
#define is_dnode_physical(addr)												\
				((__is_valid_data_blkaddr(addr) && ((addr) != ZERO_ADDR)) ?	\
				(get_dnode_nooffline(addr) & (ADDRS_PHYSICAL_TAG)) : (1))
#define set_dnode_physical(addr)											\
				((__is_valid_data_blkaddr(addr) && ((addr) != ZERO_ADDR)) ?	\
				(cpu_to_le32((addr) | (ADDRS_PHYSICAL_TAG))) :				\
				cpu_to_le32(addr))
#define get_dnode_physical(addr)											\
				((__is_valid_data_blkaddr(addr) && ((addr) != ZERO_ADDR)) ?	\
				((le32_to_cpu(addr) & (~(ADDRS_OFFLINE_TAG))) &				\
				(~(ADDRS_PHYSICAL_TAG))) : le32_to_cpu(addr))

/* treat ZERO/NULL/NEW/COMPRESS ADDRS as stable, return 0 */
#define is_dnode_offline(addr)												\
				((__is_valid_data_blkaddr(addr) && (addr != ZERO_ADDR)) ?	\
				(le32_to_cpu(addr) & (ADDRS_OFFLINE_TAG)) : (0))
#define set_dnode_offline(addr)												\
				((__is_valid_data_blkaddr(addr) && (addr != ZERO_ADDR)) ?	\
				(cpu_to_le32((addr) | (ADDRS_OFFLINE_TAG))) : cpu_to_le32(addr))
#define clear_dnode_offline(addr)											\
				((__is_valid_data_blkaddr(addr) && (addr != ZERO_ADDR)) ?	\
				cpu_to_le32((addr) & (~(ADDRS_OFFLINE_TAG))) :				\
				cpu_to_le32(addr))

static inline bool should_compute_fingerprint(struct address_space *mapping,
				struct writeback_control *wbc)
{
	if (mapping->host->i_ino <= F2FS_FP_INO(F2FS_M_SB(mapping)))
		return false;
	if (wbc->sync_mode == WB_SYNC_ALL)
		return false;
	return true;
}

static inline bool should_allocate_virtual_address(struct page *page,
				struct writeback_control *wbc)
{
	if (page->mapping->host->i_ino <= F2FS_FP_INO(F2FS_P_SB(page)))
		return false;
	if (wbc->sync_mode == WB_SYNC_ALL)
		return false;
	return true;
}

// 6. In-memory bucket/buffer list operations

static inline void slist_head_init(struct slist_head *list)
{
	// Initialize a singly linked list head, pointing to itself
	list->next = list;
}

static inline void slist_add(struct slist_head *new, struct slist_head *head)
{
	// Add a new entry after the head of the list
	new->next = head->next;
	head->next = new;
}

static inline void slist_add_tail(struct slist_head *new, struct slist_head *head)
{
	// Add a new entry at the end of the list
	struct slist_head *prev = head;
	while (prev->next) {
		prev = prev->next;
	}
	prev->next = new;
	new->next = head;
}

static inline void slist_del(struct slist_head *del, struct slist_head *head)
{
	// Remove a specific entry from the list
	struct slist_head *prev = head;
	while (prev->next != del) {
		prev = prev->next;
	}
	prev->next = del->next;
	del->next = NULL;
}

static inline void slist_del_first(struct slist_head *head)
{
	// Remove the first entry after the head
	head->next = head->next->next;
}

static inline int slist_empty(struct slist_head *head)
{
	// Check if the list is empty (head points to itself)
	return (head->next == head);
}

static inline int slist_singular(struct slist_head *head)
{
	// Check if the list has only one entry (next points back to head)
	return ((head->next != head) && (head->next->next == head));
}

#define slist_entry(ptr, type, member) \
	/* Retrieve the container structure from a list entry */ \
	container_of((ptr), type, member)

#define slist_first_entry(ptr, type, member) \
	/* Retrieve the first entry from the list */ \
	list_entry(((ptr)->next), type, member)

#define slist_for_each(pos, head) \
	/* Iterate over each entry in the list */ \
	for (((pos) = ((head)->next)); \
		 prefetch((pos)->next), ((pos) != (head)); \
		 ((pos) = ((pos)->next)))

// 7. Fingerprint kthread (kernel thread) management

#define DEF_FP_URGENT_TIME (30000) // Urgent fingerprint processing: 30 seconds
#define DEF_FP_MIN_SLEEP_TIME (30000) // Minimum sleep time: 30 seconds
#define DEF_FP_MAX_SLEEP_TIME (60000) // Maximum sleep time: 60 seconds
#define DEF_FP_NO_SLEEP_TIME (300000) // No sleep threshold: 5 minutes
#define DEF_IDLE_FP_PERSIST_INTERVAL (50) // Persistence interval: 50 seconds
#define DEF_IDLE_FP_STABLE_INTERVAL (100) // Stability interval: 100 seconds
#define DEF_IDLE_FP_OFFLINE_INTERVAL (200) // Offline interval: 200 seconds

struct f2dfs_fp_kthread {
	struct task_struct *fp_task; // Pointer to the kthread task structure
	wait_queue_head_t fp_wait_queue_head; // Wait queue for thread wakeup

	// Sleep time parameters for fingerprint processing
	unsigned int urgent_fp_sleep_time;
	unsigned int min_fp_sleep_time;
	unsigned int max_fp_sleep_time;
	unsigned int no_fp_sleep_time;
};

static inline void increase_fp_sleep_time(
	struct f2dfs_fp_kthread *fp_th, unsigned int *wait)
{
	// Gradually increase the sleep time within the defined limits
	unsigned int min_time = fp_th->min_fp_sleep_time;
	unsigned int max_time = fp_th->max_fp_sleep_time;

	if (*wait == fp_th->no_fp_sleep_time)
		return;

	if ((long long)*wait + (long long)min_time > (long long)max_time)
		*wait = max_time;
	else
		*wait += min_time;
}

static inline void decrease_fp_sleep_time(
	struct f2dfs_fp_kthread *fp_th, unsigned int *wait)
{
	// Gradually decrease the sleep time, with a minimum limit
	unsigned int min_time = fp_th->min_fp_sleep_time;

	if (*wait == fp_th->no_fp_sleep_time)
		*wait = fp_th->max_fp_sleep_time;

	if ((long long)*wait - (long long)min_time < (long long)min_time)
		*wait = min_time;
	else
		*wait -= min_time;
}

// 8. Behavioral conditions for deciding specific actions

// Check if a single bucket should be persisted during unmounting
static inline bool should_persist_single_bucket_umount(
				struct f2dfs_fp_bucket *bucket)
{
	int i = 0;
	bool exist = false;

	// If no valid or removed entries exist, skip persistence
	if ((bucket->remove_count == 0) && (bucket->valid_count == 0))
		return false;

	// If removed entries exist, mark for persistence
	if (bucket->remove_count > 0)
		exist = true;

	// Check each valid entry for persistence or stability tags
	for (i = 0; i < bucket->valid_count; ++i) {
		if (exist)
			break;
		if ((!fp_is_persist_tag(&bucket->entries[i])) ||
					(!fp_is_stable_tag(&bucket->entries[i])))
			exist = true;
	}

	return exist;
}

// Check if a single bucket should be persisted in normal operation
static inline bool should_persist_single_bucket_normal(
				struct f2dfs_fp_bucket *bucket)
{
	int i = 0;
	bool exist = false;

	// Skip persistence if no valid or removed entries exist
	if ((bucket->remove_count == 0) && (bucket->valid_count == 0))
		return false;

	// If removed entries exist, mark for persistence
	if (bucket->remove_count > 0)
		exist = true;

	// Check for persistence tags in valid entries
	for (i = 0; i < bucket->valid_count; ++i) {
		if (exist)
			break;
		if (!fp_is_persist_tag(&bucket->entries[i]))
			exist = true;
	}

	// Additional conditions for eviction
	if (!exist)
		return false;

	return ((bucket->valid_count + bucket->remove_count) >=
					FP_BUCKET_EVICT_START_1) ||
			((bucket->valid_count + (bucket->remove_count *
					FP_REMOVE_PER_ENTRY)) >= FP_BUCKET_EVICT_START_2);
}

// Determine if a single bucket has been idle long enough to free
static inline bool should_free_single_bucket(struct f2dfs_fp_bucket *bucket)
{
	return time_after(jiffies, (bucket->atime + DEF_IDLE_FP_BUCKET_INTERVAL));
}

// Determine if a single buffer has been idle long enough to free
static inline bool should_free_single_buffer(struct f2dfs_fp_buffer *buffer)
{
	return time_after(jiffies, (buffer->atime + DEF_IDLE_FP_BUFFER_INTERVAL));
}

// Check if stabilization should be performed based on rounds and sleep time
static inline bool could_do_stabilization(int rounds, unsigned int ms)
{
	return ((DEF_IDLE_FP_STABLE_ROUNDS > rounds) &&
					(DEF_IDLE_FP_STABLE_SLEEP_TIME > ms));
}

// Check if offline deduplication can be performed based on rounds and sleep time
static inline bool could_do_offline_deduplication(int rounds, unsigned int ms)
{
	return ((DEF_IDLE_FP_OFFLINE_ROUNDS > rounds) &&
					(DEF_IDLE_FP_OFFLINE_SLEEP_TIME > ms));
}

// Check if the cache can be freed based on rounds and sleep time
static inline bool could_do_free_cache(int rounds, unsigned int ms)
{
	return ((DEF_IDLE_FP_CACHE_ROUNDS > rounds) &&
					(DEF_IDLE_FP_CACHE_SLEEP_TIME > ms));
}

// Check if a single buffer should be persisted during unmounting
static inline bool should_persist_single_buffer_umount(
				struct f2dfs_fp_buffer *buffer)
{
	return (buffer->valid_count > 0);
}

// Evaluate if buffers should be persisted under normal operation
static inline bool should_persist_buffers_normal(struct f2fs_sb_info *sbi)
{
	struct slist_head *head;
	struct f2dfs_fp_buffer *buffer;
	int i, j, count = 0, number = 0;

	for (i = 0; i < FP_BUFFER_NUMBER; ++i) {
		head = &sbi->s_fingerprint_buffer->pointers[i];
		if (!head)
			goto skip;

		// Iterate over buffers in the list
		slist_for_each(head, &sbi->s_fingerprint_buffer->pointers[i]) {
			if (!head)
				goto skip;

			buffer = slist_entry(head, struct f2dfs_fp_buffer, b_head);
			if (!buffer)
				goto skip;

			count += buffer->valid_count;
			for (j = 0; j < buffer->valid_count; ++j)
				number += buffer->entries[j].number;
		}

skip:
		continue;
	}

	// Persist if entry or space thresholds are exceeded
	return ((count >= (2 * FPS_PER_BUFFER_BLOCK)) ||
					(number >= ((DEF_ADDRS_PER_BLOCK *
					FPS_PER_BUFFER_BLOCK) * 3 / 2)));
}

// Check if offline deduplication should start under normal conditions
static inline bool should_dedupe_offline_normal(struct f2fs_sb_info *sbi)
{
	return (offline_user_blocks(sbi) >
			min((block_t)FP_MIN_OFFLINE_SPACE_SIZE,
					min((block_t)FP_MIN_OFFLINE_PHYSICAL_SIZE,
							(block_t)FP_MIN_OFFLINE_LOGICAL_SIZE)));
}

// Check if urgent offline deduplication is needed
static inline bool should_dedupe_offline_urgent(struct f2fs_sb_info *sbi)
{
	return (offline_user_blocks(sbi) >
			min((block_t)FP_MAX_OFFLINE_SPACE_SIZE,
					min((block_t)FP_MAX_OFFLINE_PHYSICAL_SIZE,
							(block_t)FP_MAX_OFFLINE_LOGICAL_SIZE)));
}

// Check if there is enough space for persisting a single prefix tree
static inline bool has_not_enough_space_for_single_persist(
				struct f2fs_sb_info *sbi, struct f2dfs_fp_bucket *bucket)
{
	if (f2dfs_filesystem_free_space(sbi) <= FP_SINGLE_UNSTABLE_NEEDED(
				bucket->valid_count, bucket->remove_count)) {
		f2dfs_printk(sbi, KERN_INFO "F2DFS has not enough space to persist "
				"single prefix tree.");
		return true;
	}
	return false;
}

// Check if there is enough space for stabilizing a single prefix tree
static inline bool has_not_enough_space_for_single_stable(
				struct f2fs_sb_info *sbi)
{
	if (f2dfs_filesystem_free_space(sbi) <= FP_SINGLE_STABLE_NEEDED) {
		f2dfs_printk(sbi, KERN_INFO "F2DFS has not enough space to stabilize "
				"single prefix tree.");
		return true;
	}
	return false;
}

// Check if there is enough space for persisting offline journal entries
static inline bool has_not_enough_space_for_journal(struct f2fs_sb_info *sbi)
{
	struct slist_head *head;
	struct f2dfs_fp_buffer *buffer;
	int i, count = 0;

	for (i = 0; i < FP_BUFFER_NUMBER; ++i) {
		head = &sbi->s_fingerprint_buffer->pointers[i];
		if (!head)
			goto skip;

		// Count entries in the buffer
		slist_for_each(head, &sbi->s_fingerprint_buffer->pointers[i]) {
			if (!head)
				goto skip;

			buffer = slist_entry(head, struct f2dfs_fp_buffer, b_head);
			if (!buffer)
				goto skip;

			count += buffer->valid_count;
		}

skip:
		continue;
	}

	if (count == 0)
		return false;

	// Check space requirement
	if (f2dfs_filesystem_free_space(sbi) <= FP_OFFLINE_JOURNAL_NEEDED(count)) {
		f2dfs_printk(sbi, KERN_INFO "F2DFS has not enough space to persist "
				"offline journal and node bitmaps.");
		return true;
	}
	return false;
}

// 9. print

// Macros to log deduplication-related information, IO traces, and deduplication traces
#define f2dfs_dedupe_info(sbi, fmt, ...)				\
				f2dfs_printk(sbi, KERN_CONT fmt, ##__VA_ARGS__)
#define f2dfs_io_trace(sbi, fmt, ...)					\
				f2dfs_printk(sbi, KERN_CONT fmt, ##__VA_ARGS__)
#define f2dfs_dedupe_trace(sbi, fmt, ...)				\
				f2dfs_printk(sbi, KERN_CONT fmt, ##__VA_ARGS__)

// Function to calculate deduplication values based on input blocks, high and low values
static inline unsigned int f2dfs_dedupe_calculation(
				unsigned long blocks, unsigned int high, unsigned int low)
{
	unsigned int a = 0;

	// When 'high' value is greater than 0, perform different calculations based on its range
	if (high > 0) {
		if (high < (1 << 8))
			a = (unsigned int)((high << 4) + (low >> 28) + (blocks >> 28));
		else if (high < (1 << 18))
			a = (unsigned int)(high >> 6);
		else if (high < (1 << 28))
			a = (unsigned int)(high >> 16);
		else
			a = (high >> 28);
	}
	else {
		// When 'high' is 0, perform calculations based on the 'blocks' size
		if (blocks < (1 << 10))
			a = (unsigned int)(blocks << 2);
		else if (blocks < (1 << 20))
			a = (unsigned int)(blocks >> 8);
		else if (blocks < (1 << 30))
			a = (unsigned int)(blocks >> 18);
		else
			a = (blocks >> 30);
	}

	return a; // Return calculated deduplication value
}

// Function to determine deduplication unit based on blocks and high value
static inline unsigned int f2dfs_dedupe_unit(
				unsigned long blocks, unsigned int high)
{
	unsigned int a = 0;

	// When 'high' value is greater than 0, return different unit values based on its range
	if (high > 0) {
		if (high < (1 << 8))
			a = 3;
		else if (high < (1 << 18))
			a = 4;
		else if (high < (1 << 28))
			a = 5;
		else
			a = 6;
	}
	else {
		// When 'high' is 0, return different unit values based on 'blocks' size
		if (blocks < (1 << 10))
			a = 0;
		else if (blocks < (1 << 20))
			a = 1;
		else if (blocks < (1 << 30))
			a = 2;
		else
			a = 3;
	}

	return a; // Return deduplication unit
}

// Function to print deduplication information about the system
static inline void f2dfs_dedupe_info_printk(struct f2fs_sb_info *sbi)
{
	unsigned int dedupe_high = 0, dedupe_low = 0, data_blocks_count = 0,
			unique_blocks_count = 0, zero_blocks_count = 0,
			offline_blocks_count = 0;
	unsigned long logical_blocks_count = 0;
	char units[8][4] = {{"KB"}, {"MB"}, {"GB"}, {"TB"},
					{"PB"}, {"EB"}, {"ZB"}, {"YB"}};

	// If deduplication information printing is disabled, return early
	if (F2DFS_NOT_PRINT_DEDUPE_INFO)
		return;

	// Retrieve deduplication statistics
	dedupe_high = dedupe_user_blocks_high(sbi);
	dedupe_low = dedupe_user_blocks_low(sbi);
	data_blocks_count = max((unsigned int)(valid_user_blocks(sbi) -
					valid_node_count(sbi) - 1), (unsigned int)0);
	unique_blocks_count = unique_user_blocks(sbi);
	zero_blocks_count = zero_user_blocks(sbi);
	offline_blocks_count = offline_user_blocks(sbi);

	// Print deduplication information header
	f2dfs_dedupe_info(sbi, "/-----------------------------"
			"----------------------------\\");

	// Print inode numbers
	f2dfs_dedupe_info(sbi, "| Inode num:   Node(%1u), Meta(%1u), Root(%1u), "
			"GC(%1u), FP(%1u).   |",
			sbi->node_ino_num,
			sbi->meta_ino_num,
			sbi->root_ino_num,
			sbi->gc_ino_num,
			sbi->fp_ino_num);

	// Print fingerprint hash information
	f2dfs_dedupe_info(sbi, "| %21s (%5s, %2u B) used as F2DFS's FP. |",
			sbi->s_fingerprint_hash->basic_hash_info->full_name,
			sbi->s_fingerprint_hash->basic_hash_info->simple_name,
			sbi->s_fingerprint_hash->basic_hash_info->len);

	f2dfs_dedupe_info(sbi, "|-----------------------------"
			"----------------------------|");

	// Print number of valid user blocks, unique blocks, full-zero blocks, and deduplicated blocks
	f2dfs_dedupe_info(sbi, "| # of valid user blocks:   %10u >>> %4u %2s.     "
			"  |",
			data_blocks_count,
			f2dfs_dedupe_calculation((unsigned long)data_blocks_count, 0, 0),
			units[f2dfs_dedupe_unit((unsigned long)data_blocks_count, 0)]);

	f2dfs_dedupe_info(sbi, "| # of unique blocks:       %10u >>> %4u %2s.     "
			"  |",
			unique_blocks_count,
			f2dfs_dedupe_calculation((unsigned long)unique_blocks_count, 0, 0),
			units[f2dfs_dedupe_unit((unsigned long)unique_blocks_count, 0)]);

	f2dfs_dedupe_info(sbi, "| # of full-zero blocks:    %10u >>> %4u %2s.     "
			"  |",
			zero_blocks_count,
			f2dfs_dedupe_calculation((unsigned long)zero_blocks_count, 0, 0),
			units[f2dfs_dedupe_unit((unsigned long)zero_blocks_count, 0)]);

	// If high deduplication count is greater than 0, print deduplicated block information
	if (dedupe_high > 0)
		f2dfs_dedupe_info(sbi, "| # of deduplicated blocks: %10u.%02uE10 >>> "
				"%4u %2s.  |",
				dedupe_high,
				(dedupe_low / 100000000),
				f2dfs_dedupe_calculation(
					(unsigned long)0, dedupe_high, dedupe_low),
				units[f2dfs_dedupe_unit((unsigned long)0, dedupe_high)]);
	else
		f2dfs_dedupe_info(sbi, "| # of deduplicated blocks: %10u >>> %4u %2s."
				"       |",
				dedupe_low,
				f2dfs_dedupe_calculation((unsigned long)dedupe_low, 0, 0),
				units[f2dfs_dedupe_unit((unsigned long)dedupe_low, 0)]);

	// Print offline block count
	f2dfs_dedupe_info(sbi, "| # of offline blocks:      %10u >>> %4u %2s.     "
			"  |",
			offline_blocks_count,
			f2dfs_dedupe_calculation((unsigned long)offline_blocks_count, 0, 0),
			units[f2dfs_dedupe_unit((unsigned long)offline_blocks_count, 0)]);

	f2dfs_dedupe_info(sbi, "|-----------------------------"
			"----------------------------|");

	// Print logical blocks based on the presence of high deduplication count
	if (dedupe_high > 0)
		f2dfs_dedupe_info(sbi, "| # of logical blocks:     %4u %2s for all "
				"files & dirs.  |",
				f2dfs_dedupe_calculation((unsigned long)data_blocks_count,
					dedupe_high, dedupe_low),
				units[f2dfs_dedupe_unit((unsigned long)data_blocks_count,
					dedupe_high)]);
	else {
		logical_blocks_count = (unsigned long)(data_blocks_count +
						dedupe_low + zero_blocks_count);
		if ((data_blocks_count + dedupe_low) < data_blocks_count)
			f2dfs_dedupe_info(sbi, "| # of logical blocks:     %4u %2s for all "
						"files & dirs.  |",
						f2dfs_dedupe_calculation(
							(unsigned long)data_blocks_count, 1, dedupe_low),
						units[3]);
		else
			f2dfs_dedupe_info(sbi, "| # of logical blocks: %8lu %2s for all "
					"files & dirs.  |",
					f2dfs_dedupe_calculation(
						(unsigned long)logical_blocks_count, 0, 0),
					units[f2dfs_dedupe_unit(
						(unsigned long)logical_blocks_count, 0)]);
	}

	// Print physical block count after deduplication
	f2dfs_dedupe_info(sbi, "| # of physical blocks:    %4u %2s after "
			"deduplication.   |",
			f2dfs_dedupe_calculation((unsigned long)data_blocks_count, 0, 0),
			units[f2dfs_dedupe_unit((unsigned long)data_blocks_count, 0)]);

	// Print recorded blocks that are waiting for offline processing
	f2dfs_dedupe_info(sbi, "| # of recorded blocks:    %4u %2s wait to do "
			"offline.    |",
			f2dfs_dedupe_calculation((unsigned long)offline_blocks_count, 0, 0),
			units[f2dfs_dedupe_unit((unsigned long)offline_blocks_count, 0)]);

	// If deduplication is too aggressive, print warnings about space utilization
	if (((data_blocks_count + dedupe_low) < data_blocks_count) &&
				(dedupe_high > 0)) {
		f2dfs_dedupe_info(sbi, "|               F2DFS contributes too much."
				"               |");
		f2dfs_dedupe_info(sbi, "| Space utilization and deduplication ratio "
				"are ignored.  |");
	}
	else {
		// Print space reusage and deduplication percentage
		data_blocks_count = max((unsigned int)data_blocks_count,
						(unsigned int)1);
		logical_blocks_count = max((unsigned long)(data_blocks_count +
						dedupe_low + zero_blocks_count), (unsigned long)1);
		f2dfs_dedupe_info(sbi, "| Space reusage:%8lu.%02lux.  &&  Dedupe "
				"percent: %3lu%%.  |",
				((unsigned long)(logical_blocks_count / data_blocks_count)),
				((unsigned long)((logical_blocks_count % data_blocks_count) *
					100) / data_blocks_count),
				((unsigned long)(dedupe_low * 100) / logical_blocks_count));
	}

	f2dfs_dedupe_info(sbi, "\\-----------------------------"
			"----------------------------/");
}

enum {
	/*
	 * I. Read:
	 * 0: translation of virtual address
	 * 1~3: read, readahead, direct read of data block
	 */
	TRANSLATION = 0,
	READ_NORMAL,
	READAHEAD_NORMAL,
	READ_DIRECT,
	/*
	 * II. Buffered write:
	 * 4: inplace update of data block
	 * 5~7: outplace update of file/ inline/ directory data block
	 * 8: background GC of data block
	 * 9~11: outplace update or background GC of i/ direct/ indirect node block
	 */
	BUFFERED_IPU_FI,
	BUFFERED_OPU_FI,
	BUFFERED_OPU_IL,
	BUFFERED_OPU_DR,
	BUFFERED_OPU_BGGC,
	BUFFERED_OPU_IN,
	BUFFERED_OPU_DN,
	BUFFERED_OPU_IDN,
	/*
	 * III. Direct write:
	 * 12: outplace update of file data block
	 * 13: foreground GC of data block
	 * 14~16: foreground GC of i/ direct/ indirect node block
	 */
	DIRECT_OPU_FI,
	DIRECT_OPU_FGGC,
	DIRECT_OPU_FGGC_IN,
	DIRECT_OPU_FGGC_DN,
	DIRECT_OPU_FGGC_IDN,
	/*
	 * IV. Truncate:
	 * 17~19: trunctaion of data block/ node block/ whole file
	 */
	TRUNCATE_DATA,
	TRUNCATE_NODE,
	TRUNCATE_FILE,
};

// Function to print the IO trace logs based on the operation model
static inline void f2dfs_io_trace_printk(struct f2fs_sb_info *sbi, int model,
				virtual_t virtual_addr, nid_t nid, int ofs_in_node,
				block_t old_blkaddr, block_t new_blkaddr)
{
	// If IO trace printing is disabled, return immediately
	if (F2DFS_NOT_PRINT_IO_TRACE)
		return;

	// Switch statement for different IO operation models
	switch(model) {
		case TRANSLATION:
			// Trace for translation operations
			f2dfs_io_trace(sbi, "TR: %u VA ~~> %u PA",
					virtual_addr, old_blkaddr);
			break;

		case READ_NORMAL:
			// Trace for normal read operations
			f2dfs_io_trace(sbi, "R: %u VA ~~> %u PA",
					virtual_addr, old_blkaddr);
			break;

		case READAHEAD_NORMAL:
			// Trace for normal read-ahead operations
			f2dfs_io_trace(sbi, "RA: %u VA ~~> %u PA",
					virtual_addr, old_blkaddr);
			break;

		case READ_DIRECT:
			// Trace for direct read operations
			f2dfs_io_trace(sbi, "RD: %u VA ~~> %u PA",
					virtual_addr, old_blkaddr);
			break;

		case BUFFERED_IPU_FI:
			// Trace for buffered input operations (FI)
			f2dfs_io_trace(sbi, "BW(I): %u VA [ %u PA <=> %u PA ]",
					virtual_addr, old_blkaddr, new_blkaddr);
			break;

		case BUFFERED_OPU_FI:
			// Trace for buffered output operations (FI) for a file
			f2dfs_io_trace(sbi, "BW(O): %u VA (%u, %u) file [ %u PA --> "
					"%u PA ]",
					virtual_addr, nid, ofs_in_node, old_blkaddr, new_blkaddr);
			break;

		case BUFFERED_OPU_IL:
			// Trace for buffered output operations (IL) for inline data
			f2dfs_io_trace(sbi, "BW(O): %u VA inline [ %u PA --> %u PA ]",
					virtual_addr, old_blkaddr, new_blkaddr);
			break;

		case BUFFERED_OPU_DR:
			// Trace for buffered output operations (DR) for directory
			f2dfs_io_trace(sbi, "BW(O): %u VA (%u, %u) dir [ %u PA --> %u PA ]",
					virtual_addr, nid, ofs_in_node, old_blkaddr, new_blkaddr);
			break;

		case BUFFERED_OPU_BGGC:
			// Trace for buffered output operations (BGGC) during garbage collection
			f2dfs_io_trace(sbi, "GC(B): %u VA [ %u PA >-> %u PA ]",
					virtual_addr, old_blkaddr, new_blkaddr);
			break;

		case BUFFERED_OPU_IN:
			// Trace for buffered output operations (OPU) for inode
			f2dfs_io_trace(sbi, "BW(O)/GC(B): %u inode [ %u PA -=> %u PA ]",
					nid, old_blkaddr, new_blkaddr);
			break;

		case BUFFERED_OPU_DN:
			// Trace for buffered output operations (OPU) for direct node
			f2dfs_io_trace(sbi, "BW(O)/GC(B): %u direct node [ %u PA -=> "
					"%u PA ]",
					nid, old_blkaddr, new_blkaddr);
			break;

		case BUFFERED_OPU_IDN:
			// Trace for buffered output operations (OPU) for indirect node
			f2dfs_io_trace(sbi, "BW(O)/GC(B): %u indirect node [ %u PA -=> "
					"%u PA ]",
					nid, old_blkaddr, new_blkaddr);
			break;

		case DIRECT_OPU_FI:
			// Trace for direct output operations (FI) for a file
			f2dfs_io_trace(sbi, "DW(O): %u VA (%u, %u) [ %u PA --> %u PA ]",
					virtual_addr, nid, ofs_in_node, old_blkaddr, new_blkaddr);
			break;

		case DIRECT_OPU_FGGC:
			// Trace for direct output operations (FGGC) during garbage collection
			f2dfs_io_trace(sbi, "GC(F): %u VA [ %u PA >-> %u PA ]",
					virtual_addr, old_blkaddr, new_blkaddr);
			break;

		case DIRECT_OPU_FGGC_IN:
			// Trace for direct output operations (FGGC) for inode during garbage collection
			f2dfs_io_trace(sbi, "DW(O)/GC(F): %u inode [ %u PA -=> %u PA ]",
					nid, old_blkaddr, new_blkaddr);
			break;

		case DIRECT_OPU_FGGC_DN:
			// Trace for direct output operations (FGGC) for direct node during garbage collection
			f2dfs_io_trace(sbi, "DW(O)/GC(F): %u direct node [ %u PA -=> "
					"%u PA ]",
					nid, old_blkaddr, new_blkaddr);
			break;

		case DIRECT_OPU_FGGC_IDN:
			// Trace for direct output operations (FGGC) for indirect node during garbage collection
			f2dfs_io_trace(sbi, "DW(O)/GC(F): %u indirect node [ %u PA -=> "
					"%u PA ]",
					nid, old_blkaddr, new_blkaddr);
			break;

		case TRUNCATE_DATA:
			// Trace for truncate operations on data block
			f2dfs_io_trace(sbi, "T(D): %u VA (%u, %u) [ %u PA ]",
					virtual_addr, nid, ofs_in_node, old_blkaddr);
			break;

		case TRUNCATE_NODE:
			// Trace for truncate operations on a node
			f2dfs_io_trace(sbi, "T(N): %u node",
					nid);
			break;

		case TRUNCATE_FILE:
			// Trace for truncate operations on a file (inode)
			f2dfs_io_trace(sbi, "T(F): %u inode",
					nid);
			break;

		default:
			// Default case for unhandled models
			break;
	}
}

// Enum for trace models related to deduplication operations
enum {
	/*
	 * I. Inline deduplication:
	 */
	INLINE_TRACE_PRINTK = 0,

	/*
	 * II. Offline deduplication:
	 */

	/*
	 * III. Fingerprint Index tree:
	 */

	/*
	 * IV. Fingerprint buffer:
	 */
};

// Function to print deduplication trace logs based on the model
static inline void f2dfs_dedupe_trace_printk(int model, nid_t nid,
				int ofs_in_node, virtual_t old_vaddr, virtual_t new_vaddr)
{
	// If deduplication trace printing is disabled, return immediately
	if (F2DFS_NOT_PRINT_DEDUPE_TRACE)
		return;

	// Switch case for different deduplication trace models
	switch(model) {
		default:
			// Default case for unhandled models
			break;
	}
}

// 10. global counter

// Increment the global deduplication block count
// This function adjusts the duplication block count and unique block count
// depending on the count passed and whether unique blocks should be decremented.
static inline void f2dfs_global_inc_dedupe_block_count(
				struct f2fs_sb_info *sbi, block_t count, bool dec_unique)
{
	spin_lock(&sbi->stat_lock); // Locking the statistics to ensure atomicity

	// Check if the addition of 'count' will overflow the duplication block count
	if (sbi->duplication_block_count_low + (block_t) count <
				sbi->duplication_block_count_low &&
				sbi->duplication_block_count_high == (block_t)-1) {
		spin_unlock(&sbi->stat_lock); // Unlock and return if no changes needed
		return;
	}
	// If the low count overflows, increment the high count
	if (sbi->duplication_block_count_low + (block_t) count <
				sbi->duplication_block_count_low) {
		sbi->duplication_block_count_high += 1;
	}
	// Increment the low count by the provided count
	sbi->duplication_block_count_low += (block_t)count;

	// Decrease the unique block count if specified
	if (dec_unique)
		sbi->unique_block_count -= (block_t)count;
	
	spin_unlock(&sbi->stat_lock); // Unlock after modifications
}

// Decrement the global deduplication block count
// This function decreases the duplication block count and optionally increases the unique block count.
static inline void f2dfs_global_dec_dedupe_block_count(
				struct f2fs_sb_info *sbi, block_t count, bool inc_unique)
{
	spin_lock(&sbi->stat_lock); // Locking the statistics to ensure atomicity

	// Check if the current count is zero; if so, handle the error case
	if (sbi->duplication_block_count_low == 0 &&
				sbi->duplication_block_count_high == 0) {
		f2fs_bug_on(sbi, count != 0); // Trigger a bug if count is non-zero
		spin_unlock(&sbi->stat_lock); // Unlock and return
		return;
	}
	// Handle the case where the low count is insufficient
	if (sbi->duplication_block_count_high == 0 &&
				sbi->duplication_block_count_low < (block_t) count) {
		f2fs_bug_on(sbi, sbi->duplication_block_count_low < (block_t) count);
		spin_unlock(&sbi->stat_lock); // Unlock and return if not enough blocks
		return;
	}
	// If the low count is smaller than 'count', decrement the high count
	if (sbi->duplication_block_count_low < (block_t) count &&
				sbi->duplication_block_count_high != 0) {
		sbi->duplication_block_count_high -= 1;
	}
	// Decrease the low count by the provided count
	sbi->duplication_block_count_low -= (block_t)count;

	// Optionally increase the unique block count
	if (inc_unique)
		sbi->unique_block_count += (block_t)count;
	
	spin_unlock(&sbi->stat_lock); // Unlock after modifications
}

// Increment the valid block count globally
// This function updates the valid block count and ensures that the total valid blocks do not exceed the available user blocks.
static inline void f2dfs_global_inc_valid_block_count(
				struct f2fs_sb_info *sbi, block_t count)
{
	blkcnt_t diff = 0, release = 0;
	block_t avail_user_block_count;

	percpu_counter_add(&sbi->alloc_valid_block_count, count); // Increment the per-CPU valid block count

	spin_lock(&sbi->stat_lock); // Locking the statistics to ensure atomicity
	sbi->total_valid_block_count += (block_t)count; // Increment the total valid block count
	sbi->unique_block_count += (block_t)count; // Increment the unique block count
	avail_user_block_count = sbi->user_block_count - sbi->current_reserved_blocks;

	// Handle special case when the checkpoint is disabled
	if (unlikely(is_sbi_flag_set(sbi, SBI_CP_DISABLED))) {
		if (avail_user_block_count > sbi->unusable_block_count)
			avail_user_block_count -= sbi->unusable_block_count;
		else
			avail_user_block_count = 0;
	}

	// Check if the total valid block count exceeds the available user blocks
	if (unlikely(sbi->total_valid_block_count > avail_user_block_count)) {
		diff = sbi->total_valid_block_count - avail_user_block_count;
		if (diff > count)
			diff = count;
		count -= diff; // Adjust the count
		release = diff; // Track the released blocks
		sbi->total_valid_block_count -= diff;
		if (!count) { // If no blocks left to allocate, unlock and return
			spin_unlock(&sbi->stat_lock);
			goto enospc;
		}
	}
	spin_unlock(&sbi->stat_lock); // Unlock after modifications

	// If blocks need to be released, subtract them from the per-CPU counter
	if (unlikely(release)) {
		percpu_counter_sub(&sbi->alloc_valid_block_count, release);
	}
	return;

enospc:
	// If there is no space left, release the blocks
	percpu_counter_sub(&sbi->alloc_valid_block_count, release);
}

// Decrement the global valid block count
// This function decreases both the total valid block count and the unique block count.
static inline void f2dfs_global_dec_valid_block_count(
				struct f2fs_sb_info *sbi, block_t count)
{
	spin_lock(&sbi->stat_lock); // Locking the statistics to ensure atomicity

	// Validate the consistency of block counts
	f2fs_bug_on(sbi, sbi->total_valid_block_count < (block_t) count);
	f2fs_bug_on(sbi, sbi->unique_block_count < (block_t) count);
	f2fs_bug_on(sbi, sbi->total_valid_block_count < sbi->unique_block_count);

	// Decrease the total and unique block counts
	sbi->total_valid_block_count -= (block_t)count;
	sbi->unique_block_count -= (block_t)count;

	// Update reserved blocks if needed
	if (sbi->reserved_blocks &&
		sbi->current_reserved_blocks < sbi->reserved_blocks)
		sbi->current_reserved_blocks = min(sbi->reserved_blocks,
					sbi->current_reserved_blocks + count);
	
	spin_unlock(&sbi->stat_lock); // Unlock after modifications
}

// Increment the valid block count for a specific inode
// This function updates the valid block count for a given inode, considering quotas and available space.
static inline int f2dfs_inode_inc_valid_block_count(
				struct f2fs_sb_info *sbi, struct inode *inode,
				blkcnt_t *count)
{
	blkcnt_t diff = 0, release = 0;
	block_t avail_user_block_count, temp_total_valid_block_count;
	int ret;

	// Attempt to reserve blocks for the inode
	ret = dquot_reserve_block(inode, *count);
	if (ret)
		return ret; // Return error if unable to reserve space

	// Handle block injection for fault simulation
	if (time_to_inject(sbi, FAULT_BLOCK)) {
		f2fs_show_injection_info(sbi, FAULT_BLOCK);
		release = *count; // Track the released blocks
		goto release_quota;
	}

	spin_lock(&sbi->stat_lock); // Locking the statistics to ensure atomicity
	temp_total_valid_block_count =
					sbi->total_valid_block_count + (block_t)(*count);
	avail_user_block_count = sbi->user_block_count - sbi->current_reserved_blocks;

	// Adjust available user block count if reserved blocks are involved
	if (!__allow_reserved_blocks(sbi, inode, true))
		avail_user_block_count -= F2FS_OPTION(sbi).root_reserved_blocks;
	if (unlikely(is_sbi_flag_set(sbi, SBI_CP_DISABLED))) {
		if (avail_user_block_count > sbi->unusable_block_count)
			avail_user_block_count -= sbi->unusable_block_count;
		else
			avail_user_block_count = 0;
	}
	if (unlikely(temp_total_valid_block_count > avail_user_block_count)) {
		diff = temp_total_valid_block_count - avail_user_block_count;
		if (diff > *count)
			diff = *count;
		*count -= diff; // Adjust the count
		release = diff; // Track the released blocks
		temp_total_valid_block_count -= diff;
		if (!*count) {
			spin_unlock(&sbi->stat_lock); // Unlock if no blocks left to allocate
			goto release_quota;
		}
	}
	spin_unlock(&sbi->stat_lock); // Unlock after modifications

	// Release any reserved blocks
	if (unlikely(release)) {
		dquot_release_reservation_block(inode, release);
	}
	// Update the inode's written block count
	f2fs_i_blocks_write(inode, *count, true, true);
	return 0;

release_quota:
	// Release the reserved blocks if there's no space
	dquot_release_reservation_block(inode, release);
	return -ENOSPC; // Return out of space error
}

// Decreases the valid block count for a specific inode by a given number of blocks.
static inline void f2dfs_inode_dec_valid_block_count(
				struct f2fs_sb_info *sbi, struct inode *inode,
				block_t count)
{
	blkcnt_t sectors = count << F2FS_LOG_SECTORS_PER_BLOCK;

	// Check for inconsistencies in the inode block count
	if (unlikely(inode->i_blocks < sectors)) {
		f2fs_warn(sbi, "Inconsistent i_blocks, ino:%lu, iblocks:%llu, "
				"sectors:%llu",
				inode->i_ino, (unsigned long long)inode->i_blocks,
				(unsigned long long)sectors);
		set_sbi_flag(sbi, SBI_NEED_FSCK);
		return;
	}
	// Update the inode's block count
	f2fs_i_blocks_write(inode, count, false, true);
}

// Increases the global count of zeroed blocks.
static inline void f2dfs_global_change_zero_block_count(
				struct f2fs_sb_info *sbi, block_t count)
{
	spin_lock(&sbi->stat_lock);
	sbi->zero_block_count += (block_t)count;
	spin_unlock(&sbi->stat_lock);
}

// Updates the global count of offline blocks. If mode is 0, it resets the count.
static inline void f2dfs_global_change_offline_block_count(
				struct f2fs_sb_info *sbi, block_t count, int mode)
{
	spin_lock(&sbi->stat_lock);
	sbi->offline_block_count += (block_t)count * (mode);
	// Reset the count if mode is 0
	if (mode == 0)
		sbi->offline_block_count = 0;
	spin_unlock(&sbi->stat_lock);
}

// Structure holding the changes to various block counts.
struct f2dfs_global_counter {
	int valid_rc_change;   // Change in valid block count
	int unique_rc_change;  // Change in unique block count
	int dupl_rc_change;    // Change in duplicated block count
	int zero_rc_change;    // Change in zeroed block count
};

// Inline function to apply block count changes (valid, unique, duplicate, zero).
static inline bool f2dfs_change_block_count_inline(struct f2fs_sb_info *sbi,
				struct f2dfs_global_counter *rc_changes)
{
	/* v+-; u+-; d+-; z+- */
	blkcnt_t diff = 0, release = 0;
	block_t avail_user_block_count;
	bool valid_inc = false, valid_dec = false;
	block_t valid_rc_change = rc_changes->valid_rc_change,
					unique_rc_change = rc_changes->unique_rc_change,
					dupl_rc_change = rc_changes->dupl_rc_change,
					zero_rc_change = rc_changes->zero_rc_change;

	// If all changes are zero, return true immediately.
	if (((block_t) valid_rc_change == (block_t) 0) &&
				((block_t) unique_rc_change == (block_t) 0) &&
				((block_t) dupl_rc_change == (block_t) 0) &&
				((block_t) zero_rc_change == (block_t) 0))
		return true;

	// Determine if valid block count should be increased or decreased.
	if ((int) valid_rc_change > (int) 0)
		valid_inc = true;
	else if ((int) valid_rc_change < (int) 0)
		valid_dec = true;

	// Apply valid block count increase.
	if (valid_inc)
		percpu_counter_add(&sbi->alloc_valid_block_count, valid_rc_change);

	spin_lock(&sbi->stat_lock);

	// Check if total valid block count will go negative (invalid operation).
	if (valid_dec && ((long long) (sbi->total_valid_block_count +
				(block_t) valid_rc_change) < (long long) 0)) {
		f2fs_bug_on(sbi, (long long) (sbi->total_valid_block_count +
						(block_t) valid_rc_change) < (long long) 0);
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	// Ensure unique block count doesn't go negative.
	if (((long long) (sbi->unique_block_count + (block_t) unique_rc_change) <
				(long long) 0) && ((int) unique_rc_change < (int) 0)) {
		f2fs_bug_on(sbi, ((long long) (sbi->unique_block_count +
						(block_t) unique_rc_change) < (long long) 0) &&
						((int) unique_rc_change < (int) 0));
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	// Ensure zero block count doesn't go negative.
	if (((long long) (sbi->zero_block_count + (block_t) zero_rc_change) < 
				(long long) 0) && ((int) zero_rc_change < (int) 0)) {
		f2fs_bug_on(sbi, ((long long) (sbi->zero_block_count +
						(block_t) zero_rc_change) < (long long) 0) &&
						((int) zero_rc_change < (int) 0));
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	// If valid block count is being adjusted, ensure the total valid blocks
	// do not exceed the available user block count.
	if (valid_inc || valid_dec) {
		if (sbi->total_valid_block_count < sbi->unique_block_count) {
			f2fs_bug_on(sbi, sbi->total_valid_block_count <
							sbi->unique_block_count);
			spin_unlock(&sbi->stat_lock);
			goto enospc;
		}
	}

	// Apply valid block count increase.
	if (valid_inc) {
		sbi->total_valid_block_count += (block_t) valid_rc_change;
		avail_user_block_count = (blkcnt_t) (sbi->user_block_count - 
						sbi->current_reserved_blocks);

		// Check if there is sufficient space available for the operation.
		if (unlikely(is_sbi_flag_set(sbi, SBI_CP_DISABLED))) {
			if ((unsigned long long) avail_user_block_count >
						(unsigned long long) sbi->unusable_block_count)
				avail_user_block_count -= (blkcnt_t) sbi->unusable_block_count;
			else
				avail_user_block_count = (blkcnt_t) 0;
		}

		if (unlikely((unsigned long long) sbi->total_valid_block_count >
					(unsigned long long) avail_user_block_count)) {
			diff = (blkcnt_t) (sbi->total_valid_block_count - 
					(block_t) avail_user_block_count);
			if ((unsigned long long) diff >
						(unsigned long long) valid_rc_change)
				diff = (blkcnt_t) valid_rc_change;

			valid_rc_change -= (block_t) diff;
			release = (blkcnt_t) diff;
			sbi->total_valid_block_count -= (block_t) diff;

			if (!valid_rc_change) {
				spin_unlock(&sbi->stat_lock);
				goto enospc;
			}
		}
	}
	// Apply valid block count decrease.
	else if (valid_dec) {
		sbi->total_valid_block_count += (block_t) valid_rc_change;
		if ((sbi->reserved_blocks != (block_t) 0) &&
					((unsigned int) sbi->current_reserved_blocks <
					(unsigned int) sbi->reserved_blocks))
			sbi->current_reserved_blocks = min(
							(block_t) sbi->reserved_blocks,
							(block_t) (sbi->current_reserved_blocks - 
							(block_t) valid_rc_change));
	}

	// Apply changes to unique and zero block counts.
	sbi->unique_block_count += (block_t) unique_rc_change;
	sbi->zero_block_count += (block_t) zero_rc_change;

	// Apply changes to duplication block count.
	if ((int) dupl_rc_change > (int) 0) {
		if (((unsigned int) (sbi->duplication_block_count_low +
					(block_t) dupl_rc_change) <
					(unsigned int) sbi->duplication_block_count_low) &&
					(sbi->duplication_block_count_high == (block_t) -1)) {
			f2fs_bug_on(sbi, sbi->duplication_block_count_high ==
							(block_t) -1);
			spin_unlock(&sbi->stat_lock);
			return false;
		}

		if ((unsigned int) (sbi->duplication_block_count_low +
					(block_t) dupl_rc_change) <
					(unsigned int) sbi->duplication_block_count_low)
			sbi->duplication_block_count_high += (block_t) 1;
	}
	else if ((int) dupl_rc_change < (int) 0) {
		if (((long long) (sbi->duplication_block_count_low +
					(block_t) dupl_rc_change) < (long long) 0) &&
					(sbi->duplication_block_count_high == (block_t) 0)) {
			f2fs_bug_on(sbi, sbi->duplication_block_count_high ==
							(block_t) 0);
			spin_unlock(&sbi->stat_lock);
			return false;
		}

		if ((long long) (sbi->duplication_block_count_low +
					(block_t) dupl_rc_change) < (long long) 0)
			sbi->duplication_block_count_high -= (block_t) 1;
	}

	// Update duplication block count.
	if ((int) dupl_rc_change < (int) 0) {
		sbi->duplication_block_count_low = (block_t) 
					(sbi->duplication_block_count_low - 
					(block_t) dupl_rc_change);
	}

	spin_unlock(&sbi->stat_lock);

	// Return false if no space is available.
enospc:
	return false;
}

static inline bool f2dfs_change_block_count_direct(struct f2fs_sb_info *sbi,
				struct f2dfs_global_counter *rc_changes)
{
	/* v+; u+; d-; z- */
	blkcnt_t diff = 0, release = 0;
	block_t avail_user_block_count;
	bool valid_inc = false;
	block_t valid_rc_change = rc_changes->valid_rc_change,
					unique_rc_change = rc_changes->unique_rc_change,
					dupl_rc_change = rc_changes->dupl_rc_change,
					zero_rc_change = rc_changes->zero_rc_change;

	// Check if all the changes are zero, return true if no change is needed
	if (((block_t) valid_rc_change == (block_t) 0) &&
				((block_t) unique_rc_change == (block_t) 0) &&
				((block_t) dupl_rc_change == (block_t) 0) &&
				((block_t) zero_rc_change == (block_t) 0))
		return true;

	// Ensure there are no negative values for valid, unique, and zero block changes
	if ((int) valid_rc_change < (int) 0) {
		f2fs_bug_on(sbi, (int) valid_rc_change < (int) 0);
		return false;
	}

	if ((int) unique_rc_change < (int) 0) {
		f2fs_bug_on(sbi, (int) unique_rc_change < (int) 0);
		return false;
	}

	// Ensure that duplication and zero block changes are not positive
	if ((int) dupl_rc_change > (int) 0) {
		f2fs_bug_on(sbi, (int) dupl_rc_change > (int) 0);
		return false;
	}

	if ((int) zero_rc_change > (int) 0) {
		f2fs_bug_on(sbi, (int) zero_rc_change > (int) 0);
		return false;
	}

	// If there is a valid block change, mark it as an increment
	if ((int) valid_rc_change > (int) 0)
		valid_inc = true;

	// Update the valid block count if necessary
	if (valid_inc)
		percpu_counter_add(&sbi->alloc_valid_block_count, valid_rc_change);

	spin_lock(&sbi->stat_lock);

	// Check for overflow in zero block count
	if (((long long) (sbi->zero_block_count + (block_t) zero_rc_change) < 
				(long long) 0) && ((int) zero_rc_change < (int) 0)) {
		f2fs_bug_on(sbi, ((long long) (sbi->zero_block_count +
						(block_t) zero_rc_change) < (long long) 0)  &&
						((int) zero_rc_change < (int) 0));
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	// Validate that the total valid block count is greater than or equal to the unique block count
	if (sbi->total_valid_block_count < sbi->unique_block_count) {
		f2fs_bug_on(sbi, sbi->total_valid_block_count < sbi->unique_block_count);
		spin_unlock(&sbi->stat_lock);
		goto enospc;
	}

	// Adjust the total valid block count if necessary
	if (valid_inc) {
		sbi->total_valid_block_count += (block_t) valid_rc_change;
		avail_user_block_count = (blkcnt_t) (sbi->user_block_count - sbi->current_reserved_blocks);

		// If checkpointing is disabled, adjust available space accordingly
		if (unlikely(is_sbi_flag_set(sbi, SBI_CP_DISABLED))) {
			if ((unsigned long long) avail_user_block_count >
						(unsigned long long) sbi->unusable_block_count)
				avail_user_block_count -= (blkcnt_t) sbi->unusable_block_count;
			else
				avail_user_block_count = (blkcnt_t) 0;
		}

		// Ensure that the total valid block count does not exceed available space
		if (unlikely((unsigned long long) sbi->total_valid_block_count >
					(unsigned long long) avail_user_block_count)) {
			diff = (blkcnt_t) (sbi->total_valid_block_count -
					(block_t) avail_user_block_count);
			if ((unsigned long long) diff >
						(unsigned long long) valid_rc_change)
				diff = (blkcnt_t) valid_rc_change;

			valid_rc_change -= (block_t) diff;
			release = (blkcnt_t) diff;
			sbi->total_valid_block_count -= (block_t) diff;

			// If no valid block change remains, return false due to space shortage
			if (!valid_rc_change) {
				spin_unlock(&sbi->stat_lock);
				goto enospc;
			}
		}
	}

	// Update the unique and zero block counts
	sbi->unique_block_count += (block_t) unique_rc_change;
	sbi->zero_block_count += (block_t) zero_rc_change;

	// Check for underflow in duplication block count
	if (((long long) (sbi->duplication_block_count_low +
				(block_t) dupl_rc_change) < (long long) 0) &&
				(sbi->duplication_block_count_high == (block_t) 0)) {
		f2fs_bug_on(sbi, sbi->duplication_block_count_high == (block_t) 0);
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	// Adjust the duplication block count and handle underflow if necessary
	if ((long long) (sbi->duplication_block_count_low +
				(block_t) dupl_rc_change) < (long long) 0)
		sbi->duplication_block_count_high -= (block_t) 1;

	// Update the duplication block count low
	sbi->duplication_block_count_low += (block_t) dupl_rc_change;

	spin_unlock(&sbi->stat_lock);

	// If any valid block count needs to be released, do so
	if (valid_inc) {
		if (unlikely(release)) {
			percpu_counter_sub(&sbi->alloc_valid_block_count, release);
			return false;
		}
	}

	return true;

enospc:
	// If there's no space left, release valid blocks and return false
	if (valid_inc)
		percpu_counter_sub(&sbi->alloc_valid_block_count, release);
	return false;
}

static inline bool f2dfs_change_block_count_truncate(struct f2fs_sb_info *sbi,
				struct f2fs_global_counter *rc_changes)
{
	/* v-; u+-; d-; z- */
	block_t valid_rc_change = rc_changes->valid_rc_change,
					unique_rc_change = rc_changes->unique_rc_change,
					dupl_rc_change = rc_changes->dupl_rc_change,
					zero_rc_change = rc_changes->zero_rc_change;

	// Check if all the changes are zero, return true if no change is needed
	if (((block_t) valid_rc_change == (block_t) 0) &&
				((block_t) unique_rc_change == (block_t) 0) &&
				((block_t) dupl_rc_change == (block_t) 0) &&
				((block_t) zero_rc_change == (block_t) 0))
		return true;

	// Ensure there are no positive values for valid, duplication, and zero block changes
	if ((int) valid_rc_change > (int) 0) {
		f2fs_bug_on(sbi, (int) valid_rc_change > (int) 0);
		return false;
	}

	if ((int) dupl_rc_change > (int) 0) {
		f2fs_bug_on(sbi, (int) dupl_rc_change > (int) 0);
		return false;
	}

	if ((int) zero_rc_change > (int) 0) {
		f2fs_bug_on(sbi, (int) zero_rc_change > (int) 0);
		return false;
	}

	spin_lock(&sbi->stat_lock);

	// Ensure there is no underflow in the valid block count after change
	if ((long long) (sbi->total_valid_block_count +
				(block_t) valid_rc_change) < (long long) 0) {
		f2fs_bug_on(sbi, (long long) (sbi->total_valid_block_count +
						(block_t) valid_rc_change) < (long long) 0);
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	// Ensure there is no underflow in the unique block count
	if (((long long) (sbi->unique_block_count + (block_t) unique_rc_change) <
				(long long) 0) && ((int) unique_rc_change < (int) 0)) {
		f2fs_bug_on(sbi, ((long long) (sbi->unique_block_count +
						(block_t) unique_rc_change) < (long long) 0) &&
						((int) unique_rc_change < (int) 0));
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	// Ensure there is no underflow in the zero block count
	if ((long long) (sbi->zero_block_count + (block_t) zero_rc_change) <
				(long long) 0) {
		f2fs_bug_on(sbi, (long long) (sbi->zero_block_count +
						(block_t) zero_rc_change) < (long long) 0);
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	// Ensure the total valid block count is not less than the unique block count
	if (sbi->total_valid_block_count < sbi->unique_block_count) {
		f2fs_bug_on(sbi, sbi->total_valid_block_count < sbi->unique_block_count);
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	// Apply the changes to the block counts
	sbi->total_valid_block_count += (block_t) valid_rc_change;
	sbi->unique_block_count += (block_t) unique_rc_change;
	sbi->zero_block_count += (block_t) zero_rc_change;

	// Update reserved block counts based on changes
	if (((block_t) valid_rc_change != (block_t) 0) &&
				((sbi->reserved_blocks != (block_t) 0) &&
				((unsigned int) sbi->current_reserved_blocks <
				(unsigned int) sbi->reserved_blocks)))
		sbi->current_reserved_blocks = min((block_t) sbi->reserved_blocks,
						(block_t) ((sbi->current_reserved_blocks -
						(block_t) valid_rc_change)));

	// Ensure there is no underflow in the duplication block count
	if ((long long) (sbi->duplication_block_count_low +
				(block_t) dupl_rc_change) < (long long) 0) {
		if (sbi->duplication_block_count_high == (block_t) 0) {
			f2fs_bug_on(sbi, sbi->duplication_block_count_high ==
							(block_t) 0);
			spin_unlock(&sbi->stat_lock);
			return false;
		}

		sbi->duplication_block_count_high -= (block_t) 1;
	}

	// Apply the duplication block count change
	sbi->duplication_block_count_low += (block_t) dupl_rc_change;

	spin_unlock(&sbi->stat_lock);
	return true;
}

static inline bool f2dfs_change_block_count_offline(struct f2fs_sb_info *sbi,
				struct f2dfs_global_counter *rc_changes)
{
	/* 
	 * Initialize block change values:
	 * valid_rc_change: Change in the number of valid blocks
	 * unique_rc_change: Change in the number of unique blocks
	 * dupl_rc_change: Change in the number of duplicated blocks
	 * zero_rc_change: Change in the number of zeroed blocks
	 */
	block_t valid_rc_change = rc_changes->valid_rc_change,
					unique_rc_change = rc_changes->unique_rc_change,
					dupl_rc_change = rc_changes->dupl_rc_change,
					zero_rc_change = rc_changes->zero_rc_change;

	/* 
	 * Check if all block changes are zero. If yes, no changes are needed.
	 */
	if (((block_t) valid_rc_change == (block_t) 0) &&
				((block_t) unique_rc_change == (block_t) 0) &&
				((block_t) dupl_rc_change == (block_t) 0) &&
				((block_t) zero_rc_change == (block_t) 0))
		return true;

	/* 
	 * Validate the values of block changes:
	 * - valid_rc_change should not be positive
	 * - dupl_rc_change and zero_rc_change should not be negative
	 */
	if ((int) valid_rc_change > (int) 0) {
		f2fs_bug_on(sbi, (int) valid_rc_change > (int) 0);
		return false;
	}

	if ((int) dupl_rc_change < (int) 0) {
		f2fs_bug_on(sbi, (int) dupl_rc_change < (int) 0);
		return false;
	}

	if ((int) zero_rc_change < (int) 0) {
		f2fs_bug_on(sbi, (int) zero_rc_change < (int) 0);
		return false;
	}

	/* Ensure the sum of changes is zero */
	if ((block_t) (valid_rc_change + dupl_rc_change + zero_rc_change) !=
				(block_t) 0) {
		f2fs_bug_on(sbi, (block_t) (valid_rc_change + dupl_rc_change +
						zero_rc_change) != (block_t) 0);
		return false;
	}

	spin_lock(&sbi->stat_lock);

	/* 
	 * Ensure total valid block count does not go below zero 
	 * after applying the changes.
	 */
	if ((long long) (sbi->total_valid_block_count +
				(block_t) valid_rc_change) < (long long) 0) {
		f2fs_bug_on(sbi, (long long) (sbi->total_valid_block_count +
						(block_t) valid_rc_change) < (long long) 0);
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	/* 
	 * Ensure unique block count does not go below zero 
	 * after applying the changes.
	 */
	if ((long long) (sbi->unique_block_count + (block_t) unique_rc_change) < 
				(long long) 0) {
		f2fs_bug_on(sbi, (long long) (sbi->unique_block_count +
						(block_t) unique_rc_change) < (long long) 0);
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	/* Ensure the total valid block count is not less than the unique block count */
	if (sbi->total_valid_block_count < sbi->unique_block_count) {
		f2fs_bug_on(sbi, sbi->total_valid_block_count <
						sbi->unique_block_count);
		spin_unlock(&sbi->stat_lock);
		return false;
	}

	/* Update block counters */
	sbi->total_valid_block_count += (block_t) valid_rc_change;
	sbi->unique_block_count += (block_t) unique_rc_change;
	sbi->zero_block_count += (block_t) zero_rc_change;

	/* Adjust reserved blocks if applicable */
	if (((block_t) valid_rc_change != (block_t) 0) &&
				((sbi->reserved_blocks != (block_t) 0) &&
				((unsigned int) sbi->current_reserved_blocks <
				(unsigned int) sbi->reserved_blocks)))
		sbi->current_reserved_blocks = min((block_t) sbi->reserved_blocks,
						(block_t) (sbi->current_reserved_blocks -
						(block_t) valid_rc_change));

	/* Handle overflow of duplication block counters */
	if ((unsigned int) (sbi->duplication_block_count_low +
				(block_t) dupl_rc_change) < 
				(unsigned int) sbi->duplication_block_count_low) {
		if (sbi->duplication_block_count_high == (block_t) -1) {
			f2fs_bug_on(sbi, sbi->duplication_block_count_high == (block_t) -1);
			spin_unlock(&sbi->stat_lock);
			return false;
		}

		/* Increment high counter to handle overflow */
		sbi->duplication_block_count_high += (block_t) 1;
	}

	/* Update low duplication block count */
	sbi->duplication_block_count_low += (block_t) dupl_rc_change;

	spin_unlock(&sbi->stat_lock);
	return true;
}

/*
 * f2dfs_fingerprint.c
 */
int f2dfs_fp_is_equal(__u8 fp1[], __u8 fp2[]);
bool f2dfs_fp_is_zero(struct f2fs_sb_info *sbi, __u8 fp[], unsigned int prefix);
bool f2dfs_fp_is_null(struct f2fs_sb_info *sbi, __u8 fp[]);
bool f2dfs_fp_is_remove(__u8 fp[]);
void f2dfs_get_fp_remove(__u8 fp_remove[], unsigned int fp_prefix);
int f2dfs_page_fingerprint_update(struct f2fs_sb_info *sbi,
				struct page *page, __u8 digest[]);
unsigned int f2dfs_get_fingerprint_prefix(__u8 fp[]);
nid_t get_fpfooter_pointer(struct page *p);
bool get_fpfooter_tags(struct page *p, int type);
int set_fpfooter_pointer(struct page *p, nid_t pointer);
int set_fpfooter_tags(struct page *p, int type);
int clear_fpfooter_tags(struct page *p, int type);
int f2dfs_start_fp_thread(struct f2fs_sb_info *sbi);
void f2dfs_stop_fp_thread(struct f2fs_sb_info *sbi);
void f2dfs_enqueue_single_fp_hash_work(struct f2fs_sb_info *sbi,
				int job_no, struct page *page);
void f2dfs_init_deduplication(struct f2fs_sb_info *sbi);
void f2dfs_exit_deduplication(struct f2fs_sb_info *sbi);

/*
 * f2dfs_indedupe.c
 */
struct f2dfs_fp_bucket * f2dfs_alloc_single_bucket(
				struct f2fs_sb_info *sbi, struct slist_head *head);
struct f2dfs_fp_buffer * f2dfs_alloc_single_buffer(
				struct f2fs_sb_info *sbi, struct slist_head *head);
int f2dfs_process_single_fingerprint_for_bucket_cache(struct f2fs_sb_info *sbi,
				struct f2dfs_bucket_change *bucket_info, virtual_t source_vaddr,
				struct f2dfs_global_counter *rc_changes, int mode);
int f2dfs_append_nid_into_cache(struct f2fs_sb_info *sbi,
				nid_t inode_nid, nid_t parent_nid, int rc_change);
int f2dfs_load_nid_into_cache(struct f2fs_sb_info *sbi,
				nid_t inode_nid, nid_t parent_nid, int rc_change);
void f2dfs_shrink_single_bucket(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_bucket *bucket);
bool f2dfs_ioc_sync_buckets_cache(struct f2fs_sb_info *sbi);

/*
 * f2dfs_offdedupe.c
 */
bool f2dfs_do_offline_deduplication(struct f2fs_sb_info *sbi,
				bool umount, bool ioctl, bool urgent);
void f2dfs_sync_single_buffer_foreground(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_buffer *buffer);
int f2dfs_offline_mpage_multipages_readahead(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_offline *ofi, block_t start_bidx,
				unsigned int nr_pages);
bool f2dfs_ioc_sync_buffers_cache(struct f2fs_sb_info *sbi);
bool f2dfs_check_vaddr_from_skip_offline_buffer(struct f2fs_sb_info *sbi,
				virtual_t source_vaddr, unsigned int fp_prefix);
bool f2dfs_search_vaddr_from_skip_offline_buffer(
				struct f2fs_sb_info *sbi, virtual_t source_vaddr,
				virtual_t *target_vaddr, unsigned int fp_prefix);
void f2dfs_append_vaddr_from_skip_offline_buffer(struct f2fs_sb_info *sbi,
				virtual_t source_vaddr, virtual_t target_vaddr,
				unsigned int counter, unsigned int fp_prefix);
void f2dfs_subtract_vaddr_from_skip_offline_buffer(struct f2fs_sb_info *sbi,
				virtual_t source_vaddr, unsigned int fp_prefix);

/*
 * f2dfs_persist.c
 */
virtual_t f2dfs_do_offline_for_single_fingerprint(struct f2fs_sb_info *sbi,
				__u8 fp[], nid_t parent_nid, int ofs_in_node, virtual_t vaddr,
				unsigned int index, bool *is_physical_to_virtual);
bool f2dfs_do_persistence_and_stabilization(struct f2fs_sb_info *sbi,
				bool umount, bool ioctl, bool need_stable);
void f2dfs_sync_single_bucket_foreground(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_bucket *bucket, unsigned int index);

/*
 * f2dfs_special_inodes.c
 */
struct page *f2dfs_get_gc_page(struct f2fs_sb_info *sbi, virtual_t vaddr);
void f2dfs_clear_gc_page_dirty(struct f2fs_sb_info *sbi, virtual_t vaddr);
int f2dfs_move_gc_page(struct page *data_page,
				int gc_type, virtual_t virtual_addr);
void f2dfs_clear_fp_page_dirty(struct page *page);
void f2dfs_clear_fpnode_from_node_mapping(struct f2fs_sb_info *sbi, nid_t nid);
void f2dfs_ra_fp_page(struct f2fs_sb_info *sbi, nid_t nid);
struct page *f2dfs_get_fp_page(struct f2fs_sb_info *sbi, nid_t nid);
void f2dfs_ra_gc_page(struct f2fs_sb_info *sbi, virtual_t virtual_addr);
int f2dfs_move_fp_page(struct page *fp_page, bool umount);
int f2dfs_sync_gc_pages(struct f2fs_sb_info *sbi,
				struct writeback_control *wbc, enum iostat_type io_type);
int f2dfs_sync_fp_pages(struct f2fs_sb_info *sbi, struct writeback_control *wbc,
				bool do_balance, enum iostat_type io_type);
struct page *f2dfs_create_new_fp_page(struct dnode_of_data *dn);
struct page *f2dfs_get_new_fp_page(struct f2fs_sb_info *sbi);
struct page *f2dfs_get_new_fp_page_head(struct f2fs_sb_info *sbi, nid_t fp_nid);
int f2dfs_truncate_fp_page(struct f2fs_sb_info *sbi,
				struct page *page, __u8 alloc);

/*
 * f2dfs_virtual.c
 */
int f2fs_submit_page_read(struct inode *inode, struct page *page,
				block_t blkaddr, int op_flags, bool for_write);
void do_write_page(struct f2fs_summary *sum, struct f2fs_io_info *fio);
void set_node_addr(struct f2fs_sb_info *sbi, struct node_info *ni,
				block_t new_blkaddr, bool fsync_done);
virtual_t data_virtualaddr(struct inode *inode,
				struct page *node_page, unsigned int offset);
virtual_t f2fs_data_virtualaddr(struct dnode_of_data *dn);
block_t data_blkaddr(struct inode *inode, struct page *node_page,
				unsigned int offset);
block_t f2fs_data_blkaddr(struct dnode_of_data *dn);
int f2dfs_do_write_data_page(struct f2fs_io_info *fio,
				struct f2dfs_global_counter *rc_changes,
				__u8 page_fingerprint[]);
void f2dfs_outplace_write_data(struct dnode_of_data *dn,
				struct f2fs_io_info *fio, unsigned int fp_prefix,
				struct f2dfs_global_counter *rc_changes);
int f2dfs_inplace_write_data(struct f2fs_io_info *fio, virtual_t virtual_addr);
void f2dfs_data_blkaddr_write_path(struct dnode_of_data *dn,
				int *counter, unsigned int *fp_prefix);
int f2dfs_get_dnode_of_data_write_path(struct dnode_of_data *dn,
				pgoff_t index, int *counter, unsigned int *fp_prefix, int mode);
int f2dfs_get_dnode_of_data_offline_path(
				struct dnode_of_data *dn, pgoff_t index, int mode,
				unsigned int *fp_prefix, bool *is_stable);
void set_virtual_blkaddr(struct f2fs_sb_info *sbi,
				struct node_info *ni, block_t new_blkaddr,
				unsigned int fp_prefix, bool init, bool stable);
bool f2dfs_change_reference_count_inline(struct f2fs_sb_info *sbi,
				virtual_t virtual_addr, int mode,
				struct f2dfs_global_counter *rc_changes);
bool f2dfs_change_reference_count_offline_unique(struct f2fs_sb_info *sbi,
				virtual_t virtual_addr, unsigned int prefix,
				struct f2dfs_global_counter *rc_changes);
bool f2dfs_change_reference_count_offline_dupl(struct f2fs_sb_info *sbi,
				virtual_t src_vaddr, virtual_t trg_vaddr, unsigned int prefix,
				struct f2dfs_global_counter *rc_changes);
block_t f2dfs_address_translation(struct f2fs_sb_info *sbi,
				virtual_t virtual_addr);
block_t f2dfs_address_translation_offline_path(
				struct f2fs_sb_info *sbi, virtual_t virtual_addr,
				unsigned int *fp_prefix, bool *is_stable);
unsigned int get_fp_prefix_by_vaddr(struct f2fs_sb_info *sbi, virtual_t vaddr);
int f2dfs_get_virtual_address_info(struct f2fs_sb_info *sbi,
				virtual_t virtual_addr, struct node_info *ni);
void f2dfs_wait_on_virtual_block_writeback_range(struct inode *inode,
				virtual_t virtual_addr, block_t len);
int f2dfs_set_virtual_addr(struct f2fs_sb_info *sbi, struct vnode_of_data *vn);
int f2dfs_update_data_blkaddr(struct dnode_of_data *dn,
				struct vnode_of_data *vn);

#endif /* _F2DFS_FP_H_ */
